import argparse
import yaml 
import socket
import paramiko
import dns.resolver
import base64
import re
from pve_cloud.lib.inventory import *


def get_online_pve_host(pve_inventory, cloud_domain, target_pve):
  for pve_cluster in pve_inventory:
    if pve_cluster + "." + cloud_domain == target_pve:
      for pve_host in pve_inventory[pve_cluster]:
        # check if host is available
        pve_host_ip = pve_inventory[pve_cluster][pve_host]["ansible_host"]
        try:
            with socket.create_connection((pve_host_ip, 22), timeout=3):
                return pve_host_ip
        except Exception as e:
            # debug
            print(e, type(e))
            pass
  
  raise Exception(f"Could not find online pve host for {target_pve}")


def get_cloud_env(pve_host):
  ssh = paramiko.SSHClient()
  ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

  ssh.connect(pve_host, username="root")

  # since we need root we cant use sftp and root via ssh is disabled
  _, stdout, _ = ssh.exec_command("cat /etc/pve/cloud/cluster_vars.yaml")

  cluster_vars = yaml.safe_load(stdout.read().decode('utf-8'))

  _, stdout, _ = ssh.exec_command("cat /etc/pve/cloud/secrets/patroni.pass")

  patroni_pass = stdout.read().decode('utf-8').strip()

  # fetch bind update key for ingress dns validation
  _, stdout, _ = ssh.exec_command("sudo cat /etc/pve/cloud/secrets/internal.key")
  bind_key_file = stdout.read().decode('utf-8')

  bind_internal_key = re.search(r'secret\s+"([^"]+)";', bind_key_file).group(1)

  return cluster_vars, patroni_pass, bind_internal_key


def get_online_pve_host_prsr(args):
  print(f"export PVE_ANSIBLE_HOST='{get_online_pve_host(args.target_pve)}'")


def get_ssh_master_kubeconfig(cluster_vars, stack_name):
  resolver = dns.resolver.Resolver()
  resolver.nameservers = [cluster_vars['bind_master_ip'], cluster_vars['bind_slave_ip']]

  ddns_answer = resolver.resolve(f"masters-{stack_name}.{cluster_vars['pve_cloud_domain']}")
  ddns_ips = [rdata.to_text() for rdata in ddns_answer]

  if not ddns_ips:
    raise Exception("No master could be found via DNS!")

  ssh = paramiko.SSHClient()
  ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

  ssh.connect(ddns_ips[0], username="admin")

  # since we need root we cant use sftp and root via ssh is disabled
  _, stdout, _ = ssh.exec_command("sudo cat /etc/kubernetes/admin.conf")

  return stdout.read().decode('utf-8').replace("https://127.0.0.1:6443", f"https://{ddns_ips[0]}:6443")


def export_envr(args):
  cloud_domain = get_cloud_domain(args.target_pve)
  pve_inventory = get_pve_inventory(cloud_domain)
  ansible_host = get_online_pve_host(pve_inventory, cloud_domain, args.target_pve)
  
  cluster_vars, patroni_pass, bind_internal_key = get_cloud_env(ansible_host)
  print(f"export PVE_ANSIBLE_HOST='{ansible_host}'")
  print(f"export PVE_CLOUD_DOMAIN='{cloud_domain}'")

  # tf vars
  print(f"export PG_CONN_STR=\"postgres://postgres:{patroni_pass}@{cluster_vars['pve_haproxy_floating_ip_internal']}:5000/tf_states?sslmode=disable\"")
  print(f"export TF_VAR_pve_cloud_domain='{cloud_domain}'")
  print(f"export TF_VAR_pve_host='{ansible_host}'")
  print(f"export TF_VAR_cluster_proxy_ip='{cluster_vars['pve_haproxy_floating_ip_internal']}'")
  print(f"export TF_VAR_pve_cloud_pg_cstr=\"postgresql+psycopg2://postgres:{patroni_pass}@{cluster_vars['pve_haproxy_floating_ip_internal']}:5000/pve_cloud?sslmode=disable\"")
  print(f"export TF_VAR_master_b64_kubeconf='{base64.b64encode(get_ssh_master_kubeconfig(cluster_vars, args.stack_name).encode('utf-8')).decode('utf-8')}'")
  print(f"export TF_VAR_bind_master_ip='{cluster_vars['bind_master_ip']}'")
  print(f"export TF_VAR_bind_internal_key='{bind_internal_key}'")

  # pve inventory gets converted as b64 yaml string
  pve_64 = yaml.safe_dump(pve_inventory)
  print(f"export TF_VAR_pve_inventory_b64='{base64.b64encode(pve_64.encode('utf-8')).decode('utf-8')}'")

  
def main():
  parser = argparse.ArgumentParser(description="PVE Cloud utility cli. Should be called with bash eval.")

  base_parser = argparse.ArgumentParser(add_help=False)

  subparsers = parser.add_subparsers(dest="command", required=True)

  get_cld_domain_parser = subparsers.add_parser("get-cloud-domain", help="Get the cloud domain of a pve cluster.", parents=[base_parser])
  get_cld_domain_parser.add_argument("--target-pve", type=str, help="The target pve cluster to get the cloud domain of.", required=True)
  get_cld_domain_parser .set_defaults(func=get_cld_domain_prsr)

  export_envr_parser = subparsers.add_parser("export-envrc", help="Export variables for k8s .envrc", parents=[base_parser])
  export_envr_parser.add_argument("--target-pve", type=str, help="The target pve cluster.", required=True)
  export_envr_parser.add_argument("--stack-name", type=str, help="Stack name of the deployment.", required=True)
  export_envr_parser.set_defaults(func=export_envr)

  get_online_pve_host_parser = subparsers.add_parser("get-online-host", help="Gets the ip for the first online proxmox host in the cluster.", parents=[base_parser])
  get_online_pve_host_parser.add_argument("--target-pve", type=str, help="The target pve cluster to get the first online ip of.", required=True)
  get_online_pve_host_parser.set_defaults(func=get_online_pve_host_prsr)

  args = parser.parse_args()
  args.func(args)


if __name__ == "__main__":
  main()


import paramiko
import yaml
import pve_cloud._version


def raise_on_py_cloud_missmatch(proxmox_host):
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh.connect(proxmox_host, username="root")

    # since we need root we cant use sftp and root via ssh is disabled
    _, stdout, _ = ssh.exec_command("cat /etc/pve/cloud/cluster_vars.yaml")

    cluster_vars = yaml.safe_load(stdout.read().decode('utf-8'))

    if pve_cloud._version.__version__.startswith("0.0."): # tddog development version, skip raising
        return
  
    if cluster_vars["py_pve_cloud_version"] != pve_cloud._version.__version__:
        raise RuntimeError(f"Version missmatch! py_pve_cloud_version for cluster is {cluster_vars["py_pve_cloud_version"]}, while you are using {pve_cloud._version.__version__}")


import subprocess
from proxmoxer import ProxmoxAPI
from pve_cloud.lib.validate import raise_on_py_cloud_missmatch

def get_cloud_domain(target_pve):
    avahi_disc = subprocess.run(["avahi-browse", "-rpt", "_pxc._tcp"], stdout=subprocess.PIPE, text=True, check=True)
    services = avahi_disc.stdout.splitlines()

    # find cloud domain hosts and get first online per proxmox cluster
    for service in services:
        if service.startswith("="):
            # avahi service def
            svc_args = service.split(";")
            host_ip = svc_args[7]

            cloud_domain = None
            cluster_name = None

            for txt_arg in svc_args[9].split():
                txt_arg = txt_arg.replace('"', '')
                if txt_arg.startswith("cloud_domain"):
                    cloud_domain = txt_arg.split("=")[1]

                if txt_arg.startswith("cluster_name"):
                    cluster_name = txt_arg.split("=")[1]

            if not cloud_domain or not cluster_name:
                raise ValueError(f"Missconfigured proxmox cloud avahi service: {service}")
            
            if target_pve.endswith(cloud_domain):
                return cloud_domain
            

    raise RuntimeError("Could not get cloud domain via avahi mdns!")


def get_pve_inventory(pve_cloud_domain, skip_py_cloud_validation = False):
    # call avahi-browse -rpt _pxc._tcp and find online host matching pve cloud domain
    # connect via ssh and fetch all other hosts via proxmox api => build inventory
    avahi_disc = subprocess.run(["avahi-browse", "-rpt", "_pxc._tcp"], stdout=subprocess.PIPE, text=True, check=True)
    services = avahi_disc.stdout.splitlines()

    pve_inventory = {}

    # find cloud domain hosts and get first online per proxmox cluster
    cloud_domain_first_hosts = {}
    for service in services:
        if service.startswith("="):
            # avahi service def
            svc_args = service.split(";")
            host_ip = svc_args[7]

            cloud_domain = None
            cluster_name = None

            for txt_arg in svc_args[9].split():
                txt_arg = txt_arg.replace('"', '')
                if txt_arg.startswith("cloud_domain"):
                    cloud_domain = txt_arg.split("=")[1]

                if txt_arg.startswith("cluster_name"):
                    cluster_name = txt_arg.split("=")[1]

            if not cloud_domain or not cluster_name:
                raise ValueError(f"Missconfigured proxmox cloud avahi service: {service}")
            
            # main pve cloud inventory
            if cloud_domain == pve_cloud_domain and cluster_name not in cloud_domain_first_hosts:
                if not skip_py_cloud_validation:
                    raise_on_py_cloud_missmatch(host_ip) # validate that versions of dev machine and running on cluster match

                cloud_domain_first_hosts[cluster_name] = host_ip
            
    # iterate over hosts and build pve inv via proxmox api
    for cluster_first, first_host in cloud_domain_first_hosts.items():
        proxmox = ProxmoxAPI(
            first_host, user="root", backend='ssh_paramiko'
        )

        cluster_name = None
        status_resp = proxmox.cluster.status.get()
        for entry in status_resp:
            if entry['id'] == "cluster":
                cluster_name = entry['name']
            break

        if cluster_name is None:
            raise RuntimeError("Could not get cluster name")
        
        if cluster_name != cluster_first:
            raise ValueError(f"Proxmox cluster name missconfigured in avahi service {cluster_name}/{cluster_first}")
        
        pve_inventory[cluster_name] = {}

        # fetch other hosts via api
        cluster_hosts = proxmox.nodes.get()

        for node in cluster_hosts:
            node_name = node["node"]

            if node["status"] == "offline":
                print(f"skipping offline node {node_name}")
                continue
            
            # get the main ip
            ifaces = proxmox.nodes(node_name).network.get()
            node_ip_address = None
            for iface in ifaces:
                if 'gateway' in iface:
                    if node_ip_address is not None:
                        raise RuntimeError(f"found multiple ifaces with gateways for node {node_name}")
                    node_ip_address = iface.get("address")

            if node_ip_address is None:
                raise RuntimeError(f"Could not find ip for node {node_name}")
            
            pve_inventory[cluster_name][node_name] = {
                "ansible_user": "root",
                "ansible_host": node_ip_address
            }
                    

    return pve_inventory


import argparse
import yaml
from proxmoxer import ProxmoxAPI
import os
import paramiko
from pve_cloud.cli.pvclu import get_ssh_master_kubeconfig
from pve_cloud.lib.inventory import *


def print_kubeconfig(args):
  if not os.path.exists(args.inventory):
    print("The specified inventory file does not exist!")
    return

  with open(args.inventory, "r") as f:
    inventory = yaml.safe_load(f)

  target_pve = inventory["target_pve"]

  target_cloud_domain = get_cloud_domain(target_pve)
  pve_inventory = get_pve_inventory(target_cloud_domain)

  # find target cluster in loaded inventory
  target_cluster = None

  for cluster in pve_inventory:
    if target_pve.endswith((cluster + "." + target_cloud_domain)):
      target_cluster = cluster
      break

  if not target_cluster:
    print("could not find target cluster in pve inventory!")
    return

  first_host = list(pve_inventory[target_cluster].keys())[0]

  # connect to the first pve host in the dyn inv, assumes they are all online
  ssh = paramiko.SSHClient()
  ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
  ssh.connect(pve_inventory[target_cluster][first_host]["ansible_host"], username="root")

  # since we need root we cant use sftp and root via ssh is disabled
  _, stdout, _ = ssh.exec_command("cat /etc/pve/cloud/cluster_vars.yaml")

  cluster_vars = yaml.safe_load(stdout.read().decode('utf-8'))

  print(get_ssh_master_kubeconfig(cluster_vars, inventory["stack_name"]))


def main():
  parser = argparse.ArgumentParser(description="PVE general purpose cli for setting up.")

  base_parser = argparse.ArgumentParser(add_help=False)

  subparsers = parser.add_subparsers(dest="command", required=True)

  print_kconf_parser = subparsers.add_parser("print-kubeconfig", help="Print the kubeconfig from a k8s cluster deployed with pve cloud.", parents=[base_parser])
  print_kconf_parser.add_argument("--inventory", type=str, help="PVE cloud kubespray inventory yaml file.", required=True)
  print_kconf_parser.set_defaults(func=print_kubeconfig)

  args = parser.parse_args()
  args.func(args)


if __name__ == "__main__":
  main()


"""proxy fqdn

Revision ID: 27724e407e2b
Revises: 3c95509a5de9
Create Date: 2025-09-10 16:08:17.079028

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = '27724e407e2b'
down_revision: Union[str, Sequence[str], None] = '3c95509a5de9'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    """Upgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('k8s_ext_control_plane', sa.Column('proxy_stack_fqdn', sa.String(length=253), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('k8s_ext_control_plane', 'proxy_stack_fqdn')
    # ### end Alembic commands ###
